# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class TopologyError(RuntimeError):
    """
    Base class for all topology errors.
    """
    pass


class NotFound(TopologyError):
    """
    Raised when a node or link is not found.
    """
    pass


class AlreadyExists(TopologyError):
    """
    Raised when a node or link already exists.
    """
    pass


class Inconsistent(TopologyError):
    """
    Raised when there are inconsistencies in the topology.
    """
    pass


__all__ = ['TopologyError', 'NotFound', 'AlreadyExists', 'Inconsistent']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Topology Graph port module.
"""


from copy import deepcopy
from pprintpp import pformat


class Port:
    """
    Topology Graph Port.

    :param str label: The label of the port, relative to the node.
    :param str node_id: The identifier of the node.
    :param dict metadata: The metadata of the port.
    """

    def __init__(
        self,
        label: str,
        node_id: str,
        metadata: dict = None
    ):
        self._label = label
        self._node_id = node_id
        self._metadata = {} if metadata is None else metadata
        self._identifier = Port.calc_id(node_id, label)

        # For backwards compatibility:
        self._metadata['label'] = self._label

    def __str__(self):
        return pformat(self.as_dict())

    def as_dict(self) -> dict:
        """
        Returns the port as a dictionary.
        """
        return {
            'identifier': self._identifier,
            'label': self._label,
            'node_id': self._node_id,
            'metadata': deepcopy(self._metadata)
        }

    @property
    def label(self) -> str:
        """
        Returns the readonly label of the port, relative to the node.

        :return: The label of the port.
        :rtype: str
        """
        return self._label

    @property
    def identifier(self) -> str:
        """
        Returns the readonly identifier of the port.
        """
        return self._identifier

    @property
    def metadata(self) -> dict:
        """
        Returns the readonly metadata of the port.
        """
        return deepcopy(self._metadata)

    @property
    def node_id(self) -> str:
        """
        Returns the readonly node identifier of the port.
        """
        return self._node_id

    @classmethod
    def calc_id(
        cls,
        node_id: str,
        port_label: str
    ) -> str:
        """
        Calculates the identifier of the port.

        :param str node_id: The identifier of the node.
        :param str port_label: The relative (to the node) name of the port.
        :return: The identifier of the port.
        """
        return f'{node_id}:{port_label}'


__all__ = ['Port']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from copy import deepcopy
from warnings import warn
from typing import Iterator, Tuple

from pprintpp import pformat

from .link import Link
from .node import Node
from .port import Port
from .exceptions import NotFound


class TopologyGraph:
    """
    Represents a topology as a graph.

    :param environment: A dictionary of environment variables.
    """

    def __init__(self, environment: dict = None):
        self._nodes = {}
        self._links = {}
        self._ports = {}
        self.environment = environment if environment is not None else {}

    def as_dict(self) -> dict:
        """
        Returns a dictionary representation of the topology.
        """
        return {
            'nodes': {
                node.identifier: node.as_dict() for node in self.nodes()
            },
            'links': {
                link.identifier: link.as_dict() for link in self.links()
            },
            'ports': {
                port.identifier: port.as_dict() for port in self.ports()
            },
            'environment': deepcopy(self.environment)
        }

    def __str__(self) -> str:
        return pformat(self.as_dict())

    def create_node(
        self, node_id: str, parent: Node = None, **kwargs
    ) -> Node:
        """
        Creates a new node, adds it to the topology and returns it.

        :param node_id: The id of the node to create.
        :param kwargs: The keyword arguments to pass to the node constructor.
        """
        if self.has_node(node_id):
            return self.get_node(node_id)

        node = Node(node_id, parent=parent, metadata=kwargs)
        self._nodes[node.identifier] = node
        return node

    def create_port(self, label: str, node_id: str, **kwargs) -> Port:
        """
        This method does all the following:

        - Creates a new port
        - Adds the port to the topology
        - Adds the port to the node
        - Returns the port

        If the node does not exist, then :class:`NotFound` is raised.

        :param label: The label of the port.
        :param node_id: The id of the node.
        :param kwargs: The keyword arguments to pass to the port constructor.
        """
        if self.has_port_label(node_id, label):
            return self.get_port_by_label(node_id, label)

        node = self.get_node(node_id)

        # Create the port
        port = Port(label, node_id, metadata=kwargs)

        # Add the port to the topology
        self._ports[port.identifier] = port

        # Add the port to the node
        node.add_port(port)

        return port

    def create_link(
        self,
        node1_id: str,
        port1_label: str,
        node2_id: str,
        port2_label: str,
        **kwargs
    ) -> Link:
        """
        Creates a new link and adds it to the topology.

        :param node1_id: The id of the first node.
        :param port1_label: The label of the first port.
        :param node2_id: The id of the second node.
        :param port2_label: The label of the second port.
        :param kwargs: The keyword arguments to pass to the link constructor.
        """
        link_id = Link.calc_id(
            node1_id, port1_label, node2_id, port2_label)

        if self.has_link_id(link_id):
            return self.get_link_by_id(link_id)

        # Get the nodes
        node1 = self.get_node(node1_id)
        node2 = self.get_node(node2_id)

        # Get the ports
        port1 = self.get_port_by_label(node1_id, port1_label)
        port2 = self.get_port_by_label(node2_id, port2_label)

        # Create the link
        link = Link(node1, port1, node2, port2, metadata=kwargs)

        # Add the link to the topology
        self._links[link.identifier] = link

        return link

    def get_node(self, node_id: str) -> Node:
        """
        Returns the node with the given id.

        :param node_id: The id of the node to return.
        """
        if node_id not in self._nodes:
            raise NotFound(f'Node {node_id} not found in the topology')
        return self._nodes[node_id]

    def get_port_by_id(self, port_id: str) -> Port:
        """
        Returns the port with the given id.

        :param port_id: The id of the port to return.
        """
        if port_id not in self._ports:
            raise NotFound(f'Port {port_id} not found in the topology')
        return self._ports[port_id]

    def get_port_by_label(self, node_id: str, port_label: str) -> Port:
        """
        Returns the port with the given label.

        :param node_id: The id of the node.
        :param port_label: The label of the port.
        """
        if node_id not in self._nodes:
            raise NotFound(f'Node {node_id} not found in the topology')
        return self._nodes[node_id].get_port_by_label(port_label)

    def get_link_by_id(self, link_id: str) -> Link:
        """
        Returns the link with the given id.

        :param link_id: The id of the link to return.
        """
        if link_id not in self._links:
            raise NotFound(f'Link {link_id} not found in the topology')
        return self._links[link_id]

    def get_link(
        self,
        node1_id: str,
        port1_label: str,
        node2_id: str,
        port2_label: str
    ) -> Link:
        """
        Returns the link between the given ports.

        :param node1_id: The id of the first node.
        :param port1_label: The label of the first port.
        :param node2_id: The id of the second node.
        :param port2_label: The label of the second port.
        """
        link_id = Link.calc_id(
            node1_id, port1_label, node2_id, port2_label)
        return self.get_link_by_id(link_id)

    def has_node(self, node_id: str) -> bool:
        """
        Returns True if the topology has a node with the given id.

        :param node_id: The id of the node to check.
        """
        return node_id in self._nodes

    def has_port_id(self, port_id: str) -> bool:
        """
        Returns True if the topology has a port with the given id.

        :param port_id: The identifier of the port to check.
        """
        return port_id in self._ports

    def has_port_label(self, node_id: str, port_label: str) -> bool:
        """
        Returns True if the topology has a port with the given label.

        :param node_id: The id of the node.
        :param port_label: The label of the port.
        """
        if node_id not in self._nodes:
            raise NotFound(f'Node {node_id} not found in the topology')
        return self._nodes[node_id].has_port_label(port_label)

    def has_link_id(self, link_id: str) -> bool:
        """
        Returns True if the topology has a link with the given id.

        :param link_id: The id of the link to check.
        """
        return link_id in self._links

    def has_link(
        self,
        node1_id: str,
        port1_label: str,
        node2_id: str,
        port2_label: str
    ) -> bool:
        """
        Returns True if the topology has a link between the given ports.

        :param node1_id: The id of the first node.
        :param port1_label: The label of the first port.
        :param node2_id: The id of the second node.
        :param port2_label: The label of the second port.
        """
        link_id = Link.calc_id(
            node1_id, port1_label, node2_id, port2_label)
        return self.has_link_id(link_id)

    def nodes(self) -> Iterator[Node]:
        """
        Returns an iterator over all nodes in the topology.
        """
        yield from self._nodes.values()

    def ports(self) -> Iterator[Port]:
        """
        Returns an iterator over all ports in the topology.
        """
        yield from self._ports.values()

    def links(self) -> Iterator[Link]:
        """
        Returns an iterator over all links in the topology.
        """
        yield from self._links.values()

    def bilinks(self) -> Iterator[
        Tuple[Tuple[Node, Port],
              Tuple[Node, Port], Link]
    ]:
        """
        Returns an iterator over tuples of the form:
        ((node1, port1), (node2, port2)), link). That is, a 3-tuple where the
        first two elements are 2-tuples of the form (node, port) and the third
        element is the link.

        This method exists for backwards compatibility. Use links() instead.
        """
        warn(
            'TopologyGraph.bilinks() is deprecated. Use links() instead.',
            DeprecationWarning
        )
        for link in self._links.values():
            yield ((link.node1, link.port1), (link.node2, link.port2), link)

    def biports(self) -> Iterator[Tuple[Node, Port]]:
        """
        Returns an iterator over tuples of the form:
        (node, port). That is, a 2-tuple where the first element is the node
        and the second element is the port.

        This method exists for backwards compatibility. Use ports() instead.
        """
        warn(
            'TopologyGraph.biports() is deprecated. Use ports() instead.',
            DeprecationWarning
        )
        for node in self._nodes.values():
            for port in node.ports():
                yield (node, port)

    def check_consistency(self):
        """
        Checks the consistency of the topology which includes:

        - Whenever a node has a parent, that parent must exist, and the parent
          must have the node as a subnode.
        - Whenever a node has a subnode, that subnode must exist.
        - Whenever a link has a node1, that node must exist.
        - Whenever a link has a node2, that node must exist.
        - Whenever a link has a port1, that port must exist in the topology
          and in the node1.
        - Whenever a link has a port2, that port must exist in the topology
          and in the node2.
        """

        # Validate parent/child relationships
        for node in self._nodes.values():
            for subnode in node._subnodes:
                if subnode not in self._nodes:
                    raise NotFound(
                        f'Node {subnode.identifier} which is children of '
                        f'{node.identifier} was not found in the topology')

            if (
                node.parent is not None and
                node.parent.identifier not in self._nodes
            ):
                raise NotFound(
                    f'Node {node.parent.id} which is parent of '
                    f'{self.identifier} was not found in the topology')

        # Validate link/node relationships
        for link in self._links.values():
            if link.node1 not in self.nodes():
                raise NotFound(
                    f'Node {link.node1.identifier} which is node1 of '
                    f'{link.identifier} was not found in the topology')

            if link.node2 not in self.nodes():
                raise NotFound(
                    f'Node {link.node2.identifier} which is node2 of '
                    f'{link.identifier} was not found in the topology')

            if link.port1 not in self.ports():
                raise NotFound(
                    f'Port {link.port1.identifier} which is port1 of '
                    f'{link.identifier} was not found in the topology')

            if link.port2 not in self.ports():
                raise NotFound(
                    f'Port {link.port2.identifier} which is port2 of '
                    f'{link.identifier} was not found in the topology')

            if link.port1 not in link.node1.ports():
                raise NotFound(
                    f'Port {link.port1.identifier} which is port1 of '
                    f'{link.identifier} was not found in the node '
                    f'{link.node1.identifier}')

            if link.port2 not in link.node2.ports():
                raise NotFound(
                    f'Port {link.port2.identifier} which is port2 of '
                    f'{link.identifier} was not found in the node '
                    f'{link.node2.identifier}')


__all__ = ['TopologyGraph']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Argument management module.
"""

import logging
from os import getcwd
from re import compile
from pprint import pformat
from collections import OrderedDict
from distutils.dir_util import mkpath
from argparse import Action, ArgumentParser
from os.path import join, isabs, abspath, isfile

from . import __version__
from .platforms.manager import platforms, DEFAULT_PLATFORM


log = logging.getLogger(__name__)

FORMAT = '%(asctime)s:::%(levelname)s:::%(message)s'
V_LEVELS = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}
REGEX_SLUG = compile(r'^[a-zA-Z][a-zA-Z0-9_-]*$')


class ExtendAction(Action):
    """
    Action that allows to combine nargs='+' with the action='append' behavior
    but generating a flat list.

    This allow to specify in a argument parser class an option that allows
    to specify multiple values per argument and multiple arguments.

    Usage::

        parser = ArgumentParser(...)
        parser.register('action', 'extend', ExtendAction)

        parser.add_argument(
            '-o', '--option',
            nargs='+',
            dest='options',
            action='extend',
            help='Some description'
        )

    Then, in CLI::

        executable --option var1=var1 var2=var2 --option var3=var3

    And this generates a::

        Namespace(options=['var1=var1', 'var2=var2', 'var3=var3'])
    """

    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest) or []
        items.extend(values)
        setattr(namespace, self.dest, items)


class InvalidArgument(Exception):
    """
    Typed exception that allows to fail in argument parsing and verification
    without quiting the process.
    """
    pass


def booleanize(value):
    """
    Convert a string to a boolean.

    :raises: ValueError if unable to convert.

    :param str value: String to convert.

    :return: True for 'yes' and 'true', False for 'no' and
     'false'. Not case sensitive.
    :rtype: bool
    """
    valuemap = {
        'true': True,
        'yes': True,
        'false': False,
        'no': False,
    }
    casted = valuemap.get(value.lower(), None)
    if casted is None:
        raise ValueError(str(value))
    return casted


def parse_options(raw):
    """
    Parse a list of options given by the user.

    :param list raw: Raw options given by the user in the CLI. Raw options are
     in the form:

     .. code-block:: python3

        ['var1=yes', 'var2=1.2', 'var3=40', 'var4=someoption']

    :return: Parsed option. Keys are verified safe for passing as keyword
     arguments and values are "autocasted" to a Python datatype. Declaration
     order is respected, so latest options will override the first ones. Using
     the above example, returned options will be:

     .. code-block:: python3

        OrderedDict([
            ('var1', True),
            ('var2', 1.2),
            ('var3', 40),
            ('var4', 'someoption'),
        ])

    :rtype: OrderedDict

    :raises: InvalidArgument if options have a syntax issue.
    """
    options = OrderedDict()
    if not raw:
        return options

    for option in raw:

        if '=' not in option:
            raise InvalidArgument(
                'Invalid option "{}", options must follow '
                'the syntax "<option_name>=<value>"'.format(option)
            )

        key, value = option.split('=', 1)

        # Check key form
        if not REGEX_SLUG.match(key):
            raise InvalidArgument(
                'Option "{}" is invalid. It must match "{}"'.format(
                    key, REGEX_SLUG.pattern,
                )
            )
        key = key.replace('-', '_')

        # Try to cast value
        for caster in [
            int,
            float,
            booleanize,
        ]:
            try:
                value = caster(value)
                break
            except Exception:
                continue

        options[key] = value

    return options


def validate_args(args):
    """
    Validate that arguments are valid.

    :param args: An arguments namespace.
    :type args: :py:class:`argparse.Namespace`

    :return: The validated namespace.
    :rtype: :py:class:`argparse.Namespace`
    """
    level = V_LEVELS.get(args.verbose, logging.DEBUG)
    logging.basicConfig(format=FORMAT, level=level)

    log.debug('Raw arguments:\n{}'.format(args))

    # Verify topology file exists
    if not isfile(args.topology):
        raise InvalidArgument(
            'No such file {}'.format(args.topology)
        )
    args.topology = abspath(args.topology)

    # Verify inject file exists
    if args.inject:
        if not isfile(args.inject):
            raise InvalidArgument(
                'No such file {}'.format(args.inject)
            )
        args.inject = abspath(args.inject)

    # Determine log directory and create it if required
    if args.log_dir:
        if not isabs(args.log_dir):
            args.log_dir = join(abspath(getcwd()), args.log_dir)
        mkpath(args.log_dir)

    # Parse options
    if args.options:
        args.options = parse_options(args.options)
        log.debug('Options given:\n{}'.format(pformat(args.options)))

    return args


def parse_args(argv=None):
    """
    Argument parsing routine.

    :param list argv: A list of argument strings.

    :return: A parsed and verified arguments namespace.
    :rtype: :py:class:`argparse.Namespace`
    """

    parser = ArgumentParser(
        description=(
            'Topology is a framework for building and testing network '
            'topologies, with support for pytest.'
        )
    )
    parser.register('action', 'extend', ExtendAction)

    # Standard options
    parser.add_argument(
        '-v', '--verbose',
        help='Increase verbosity level',
        default=0,
        action='count'
    )
    parser.add_argument(
        '--version',
        action='version',
        version='Topology Framework v{}'.format(__version__)
    )

    # Platforms options
    parser.add_argument(
        '--platform',
        default=DEFAULT_PLATFORM,
        help='Platform engine to build the topology with',
        choices=platforms()
    )
    parser.add_argument(
        '-o', '--option',
        nargs='+',
        dest='options',
        action='extend',
        help='Platform options'
    )

    # Ploting options
    parser.add_argument(
        '--plot-dir',
        default=None,
        help=(
            'Obsolete. No longer supported. Used in previous versions of '
            'topology to specify the directory to autoplot topologies'
        )
    )
    parser.add_argument(
        '--plot-format',
        default='svg',
        help=(
            'Obsolete. No longer supported. Used in previous versions of '
            'topology to specify the plot format'
        )
    )

    # Export options
    parser.add_argument(
        '--nml-dir',
        default=None,
        help=(
            'Obsolete. No longer supported. Used in previous versions of '
            'topology to specify the directory to export topologies as NML XML'
        )
    )

    # Logging options
    parser.add_argument(
        '--log-dir',
        default=None,
        help='Directory to create log files'
    )
    parser.add_argument(
        '--show-build-commands',
        help='Show commands executed in nodes during build',
        action='store_true'
    )

    # Building options
    parser.add_argument(
        '--inject',
        default=None,
        help='Path to an attributes injection file'
    )
    parser.add_argument(
        '--non-interactive',
        help='Just build the topology and exit',
        action='store_true'
    )
    parser.add_argument(
        'topology',
        help='File with the topology description to build'
    )

    args = parser.parse_args(argv)
    args = validate_args(args)
    return args


__all__ = [
    'InvalidArgument',
    'parse_options',
    'parse_args',
]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Topology Graph Node module.
"""


from warnings import warn
from copy import deepcopy
from typing import Iterator

from pprintpp import pformat
from typing_extensions import Self

from .port import Port
from .exceptions import AlreadyExists, Inconsistent, NotFound


class Node:
    """
    Topology Node.

    :param str identifier: The identifier of the node.
    :param Node parent: The parent node.
    :param dict metadata: The metadata of the node.
    """

    def __init__(
        self,
        identifier: str,
        parent: Self = None,
        metadata: dict = None
    ):
        self._identifier = identifier
        self._metadata = {} if metadata is None else metadata
        self._subnodes = {}
        self._ports = {}

        """
        The parent node.
        """
        self.parent = parent

    def as_dict(self) -> dict:
        """
        Returns the node as a dictionary.
        """
        return {
            'identifier': self._identifier,
            'metadata': deepcopy(self._metadata),
            'subnodes': {
                subnode_id: subnode.as_dict()
                for subnode_id, subnode in self._subnodes.items()
            },
            'ports': {
                port_id: port.as_dict()
                for port_id, port in self._ports.items()
            }
        }

    def __str__(self):
        return pformat(self.as_dict())

    @property
    def identifier(self) -> str:
        """
        Returns the readonly identifier of the node.
        """
        return self._identifier

    @property
    def name(self) -> str:
        """
        Returns the name of the node.

        This exists for backwards compatibility. It is recommended to use
        metadata instead.
        """
        warn(
            'Use of GraphNode.name is deprecated. Use metadata.get("name") '
            'instead.',
            DeprecationWarning
        )
        return self._metadata.get('name')

    @property
    def metadata(self) -> dict:
        """
        Returns a deepcopy of the metadata of the node.
        """
        return deepcopy(self._metadata)

    def has_parent(self) -> bool:
        """
        Returns True if the node has a parent.
        """
        return self.parent is not None

    def add_subnode(self, subnode: Self):
        """
        Adds a subnode to the node.

        :param Node subnode: The subnode to add.
        """
        if subnode.identifier in self._subnodes:
            raise AlreadyExists(
                f'Subnode {subnode.identifier} already exists in node '
                f'{self.identifier}'
            )
        self._subnodes[subnode.identifier] = subnode

    def has_subnode(self, subnode_id: str) -> bool:
        """
        Returns True if the node has a subnode with the given identifier.

        :param str subnode_id: The identifier of the subnode.
        """
        return subnode_id in self._subnodes

    def get_subnode(self, subnode_id: str) -> Self:
        """
        Returns the subnode with the given identifier.

        :param str subnode_id: The identifier of the subnode.
        """
        if subnode_id not in self._subnodes:
            raise NotFound(
                f'Subnode {subnode_id} not found in node '
                f'{self.identifier}'
            )
        return self._subnodes[subnode_id]

    def subnodes(self) -> Iterator[Self]:
        """
        Returns an iterator over all subnodes in the node.
        """
        yield from self._subnodes.values()

    def ports(self) -> Iterator[Port]:
        """
        Returns an iterator over all ports in the node.
        """
        yield from self._ports.values()

    def add_port(self, port: Port):
        """
        Adds a port to the node.

        :param Port port: The port to add.
        """
        if port.node_id != self.identifier:
            raise Inconsistent(
                f'Cannot add port {port.identifier} to node {self.identifier} '
                'because port.node_id points to a different '
                f'node: {port.node_id}'
            )
        if port.identifier in self._ports:
            raise AlreadyExists(
                f'Port {port.identifier} already exists in '
                f'node {self.identifier}')
        self._ports[port.identifier] = port

    def get_port_by_id(self, port_id: str) -> Port:
        """
        Returns the port with the given identifier.

        :param str port_id: The identifier of the port.
        """
        if port_id not in self._ports:
            raise NotFound(
                f'Port {port_id} not found in node {self.identifier}'
            )
        return self._ports[port_id]

    def get_port_by_label(self, port_label: str) -> Port:
        """
        Returns the port with the given label.

        :param str port_label: The label of the port.
        """
        port_id = Port.calc_id(self.identifier, port_label)
        return self.get_port_by_id(port_id)

    def has_port_id(self, port_id: str) -> bool:
        """
        Returns True if the node has a port with the given identifier.

        :param str port_id: The identifier of the port.
        """
        return port_id in self._ports

    def has_port_label(self, port_label: str) -> bool:
        """
        Returns True if the node has a port with the given label.

        :param str port_label: The label of the port.
        """
        port_id = Port.calc_id(self.identifier, port_label)
        return self.has_port_id(port_id)


__all__ = ['Node']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
topology manager module.

This module takes care of understanding the topology defined in the test
suite, commanding its construction to appropriate topology platform plugin
and finally commanding its destruction to the plugin also.
"""

from __future__ import unicode_literals, absolute_import
from __future__ import print_function, division

import logging
from sys import exc_info
from warnings import warn
from copy import deepcopy
from inspect import signature
from datetime import datetime
from traceback import format_exc
from collections import OrderedDict

from six import string_types
from pyszn.parser import parse_txtmeta

from .graph import TopologyGraph, Link
from .platforms.node import BaseNode
from .platforms.manager import platforms, load_platform, DEFAULT_PLATFORM


log = logging.getLogger(__name__)


class TopologyManager(object):
    """
    Main Topology Manager object.

    This object is responsible to build a topology given a description.

    There is three options to build a topology:

    - Using the simplified textual description using a Graphviz like syntax.
      To use this option call the :meth:`parse`.
    - Using a basic Python dictionary to load the description of the topology.
      To use this option call the :meth:`load`.
    - Set the :attr:`graph` attribute to a :class:`TopologyGraph` object. For
      example:

      ::

         from topology.graph import TopologyGraph
         from topology.manager import TopologyManager

         mgr = TopologyManager()
         graph = TopologyGraph()
         graph.add_node(Node('My Node'))
         mgr.graph = graph
         # ...

      See :class:`topology.graph.TopologyGraph` for more information of
      this usage.

    :param str engine: Name of the platform engine to build the topology.
     See :func:`platforms` for how to get and discover available platforms.
    :param dict options: Options to pass to the topology platform
    """

    def __init__(self, engine=DEFAULT_PLATFORM, options=None, **kwargs):
        super(TopologyManager, self).__init__()

        if engine not in platforms():
            raise RuntimeError('Unknown platform engine "{}".'.format(engine))

        self.graph = TopologyGraph()
        self.engine = engine
        self.options = options or OrderedDict()
        self.nodes = OrderedDict()
        self.ports = OrderedDict()

        self._platform = None
        self._built = False

    @property
    def nml(self):
        """
        Topology no longer users NML specification. This method now returns
        a :class:`topology.graph.TopologyGraph` object which was carefully
        made to have the same interface as ExtendedNMLManger used to have.

        This is obsolete and will be removed in future versions. Use the
        :attr:`graph` attribute instead.
        """
        warn(
            "nml attribute is obsolete and will be removed in future "
            "versions. Use graph attribute instead.",
            DeprecationWarning
        )
        return self.graph

    def load(self, dictmeta, inject=None):
        """
        Load a topology description in a dictionary format.

        Dictionary format:

        ::

            {
                'nodes': [
                    {
                        'nodes': ['sw1', 'hs1', ...],
                        'attributes': {...}
                    },
                ],
                'ports': [
                    {
                        'ports': [('sw1', '1'), ('hs1', '1'), ...],
                        'attributes': {...}
                    }
                ]
                'links': [
                    {
                        'endpoints': (('sw1', '1'), ('hs1', '1')),
                        'attributes': {...}
                    },
                ]
            }

        :param dict dictmeta: The dictionary to load the topology from.
        :param dict inject: An attributes injection sub-dictionary as defined
         by :func:`parse_attribute_injection`.
        """
        # Load the environment
        environment = dictmeta.get('environment', OrderedDict())
        if inject is not None and 'environment' in inject:
            environment.update(inject['environment'])

        self.graph.environment = environment

        # Load nodes
        node_to_parent = {}
        for nodes_spec in dictmeta.get('nodes', []):
            for node_id in nodes_spec['nodes']:

                # Get node attributes
                attrs = deepcopy(nodes_spec['attributes'])

                # Inject the run-specific attributes
                if inject is not None and node_id in inject['nodes']:
                    attrs.update(inject['nodes'][node_id])

                # Create a new node
                self.graph.create_node(node_id, **attrs)

                parent_id = nodes_spec['parent']
                if parent_id is not None:
                    node_to_parent[node_id] = parent_id

        # Load ports
        for ports_spec in dictmeta.get('ports', []):
            for node_id, port_label in ports_spec['ports']:

                # Create the node if still not created
                self.graph.create_node(node_id)

                # Get port attributes
                attrs = deepcopy(ports_spec['attributes'])

                # Inject the run-specific attributes
                if (
                    inject is not None and
                    (node_id, port_label) in inject['ports']
                ):
                    attrs.update(inject['ports'][(node_id, port_label)])

                # Create a new port. This call also adds the port to the
                # topology and node.
                self.graph.create_port(port_label, node_id, **attrs)

        # Load links
        for link_spec in dictmeta.get('links', []):

            # Get link attributes
            attrs = deepcopy(link_spec['attributes'])

            # Inject the run-specific attributes
            if (
                inject is not None and
                link_spec['endpoints'] in inject['links']
            ):
                attrs.update(inject['links'][link_spec['endpoints']])

            # This variable will hold these data for the link:
            # endpoints = [(node1_id, port1_label), (node2_id, port2_label)]
            endpoints = []

            for node_id, port_label in link_spec['endpoints']:

                # Create the node if it's still not created
                self.graph.create_node(node_id)

                # Create the port if it's still not created.
                self.graph.create_port(port_label, node_id)

                # Register endpoint
                endpoints.append((node_id, port_label))

            # Decompose the endpoints list into two tuples
            node1_id, port1_label = endpoints[0]
            node2_id, port2_label = endpoints[1]

            # Create a new link
            self.graph.create_link(
                node1_id, port1_label, node2_id, port2_label, **attrs
            )

        # Set parent-child relationships
        for node_id, parent_id in node_to_parent.items():
            node = self.graph.get_node(node_id)
            parent = self.graph.get_node(parent_id)

            # Set the child's parent
            node.parent = parent

            # Add the child to the parent's children
            parent.add_subnode(node)

    def parse(self, txtmeta, load=True, inject=None):
        """
        Parse a textual topology meta-description.

        For a description of the textual format see pyszn package
        documentation.

        :param str txtmeta: The textual meta-description of the topology.
        :param bool load: If ``True`` (the default) call :meth:`load`
         immediately after parse.
        :param dict inject: An attributes injection sub-dictionary as defined
         by :func:`parse_attribute_injection`.
        """
        data = parse_txtmeta(txtmeta)
        if load:
            self.load(data, inject=inject)
        return data

    def is_built(self):
        """
        Check if the current topology was built.

        :rtype: bool
        :return: True if the topology was successfully built.
        """
        return self._built

    def build(self):
        """
        Build the topology hold by this manager.

        This method instance the platform engine and request the build of the
        topology defined.
        """
        if self._built:
            raise RuntimeError(
                'You cannot build a topology twice.'
            )

        node_enode_map = OrderedDict()
        timestamp = datetime.now().replace(microsecond=0).isoformat()
        stage = 'instance'

        # Instance platform
        plugin = load_platform(self.engine)

        # Check if platform support to pass arguments to keep backwards
        # compatibility
        if any(
            param.kind == param.VAR_KEYWORD
            for param in reversed(signature(plugin).parameters.values())
        ):
            self._platform = plugin(
                timestamp, self.graph, **self.options
            )
        else:
            if self.options:
                msg = (
                    'Ignoring platform options "{}" '
                    'as current platform "{}" does not support to pass options'
                ).format(self.options, self.engine)
                log.warning(msg)
                warn(msg, DeprecationWarning)
            self._platform = plugin(timestamp, self.graph)

        try:
            stage = 'pre_build'
            self._platform.pre_build()

            stage = 'add_node'
            for node in self.graph.nodes():
                enode = self._platform.add_node(node)

                # Check that engine node implements the minimum interface
                if not isinstance(enode, BaseNode):
                    msg = (
                        'Platform {} returned an invalid '
                        'engine node {}.'
                    ).format(self.engine, enode)
                    log.critical(msg)
                    raise Exception(msg)

                # Register engine node
                node_enode_map[node.identifier] = enode.identifier
                self.nodes[enode.identifier] = enode
                # Register empty port map
                self.ports[enode.identifier] = OrderedDict()

            stage = 'add_biport'
            for node in self.graph.nodes():
                for port in node.ports():
                    eport = self._platform.add_biport(node, port)

                    # Check that engine port is of correct type
                    if not isinstance(eport, string_types):
                        msg = (
                            'Platform {} returned an invalid '
                            'engine port name {}.'
                        ).format(self.engine, enode)
                        log.critical(msg)
                        raise Exception(msg)

                    # Register engine port
                    label = port.metadata.get('label', port.identifier)
                    enode_id = node_enode_map[node.identifier]
                    self.ports[enode_id][label] = eport

            stage = 'add_bilink'
            for link in self.graph.links():
                node_porta = (link.node1, link.port1)
                node_portb = (link.node2, link.port2)
                self._platform.add_bilink(node_porta, node_portb, link)

            stage = 'post_build'

            # Assign the port mapping to the enode so they know their mapping
            # and be able to change it if required
            # (and globally, we do not use deepcopy).
            for enode_id, enode in self.nodes.items():
                enode.ports = self.ports[enode_id]

            self._platform.post_build()

        except (Exception, KeyboardInterrupt):
            e = exc_info()[1]
            log.critical(
                (
                    'Build failed at stage "{}" with "{}". '
                    'Calling plugin rollback routine...'
                ).format(stage, e)
            )
            log.debug(format_exc())
            self._platform.rollback(stage, self.nodes, e)
            raise

        self._built = True

    def unbuild(self):
        """
        Undo the topology.

        This removes all references to the engine nodes and request the
        platform to destroy the topology.
        """
        if not self._built:
            raise RuntimeError(
                'You cannot unbuild and never built topology.'
            )

        # Remove own reference to enodes
        self.nodes = OrderedDict()

        # Call platform destroy hook
        self._platform.destroy()

        # Explicitly delete platform
        del self._platform
        self._platform = None

    def get(self, identifier):
        """
        Get a platform node with given identifier.

        :param str identifier: The node identifier.
        :rtype: A subclass of :class:`topology.platforms.node.BaseNode`
        :return: The engine node implementing the communication with the node.
        """
        return self.nodes.get(identifier, None)

    def relink(self, link_id):
        """
        Relink back a link specified in the topology.

        :param str link_id: Link identifier to be recreated.
        """
        warn(
            'relink() is deprecated and will be removed in future releases.'
            'Use set_link() instead.',
            DeprecationWarning
        )
        if not self._built:
            raise RuntimeError(
                'You cannot relink on a never built topology.'
            )
        self._platform.relink(link_id)

    def unlink(self, link_id):
        """
        Unlink (break) a link specified in the topology.

        :param str link_id: Link identifier to be recreated.
        """
        warn(
            'unlink() is deprecated and will be removed in future releases.'
            'Use unset_link() instead.',
            DeprecationWarning
        )
        if not self._built:
            raise RuntimeError(
                'You cannot unlink on a never built topology.'
            )
        self._platform.unlink(link_id)

    def unset_link(self, node1_id, port1_label, node2_id, port2_label):
        """
        Unset a link between two nodes.

        :param str node1_id: The first node identifier.
        :param str port1_label: The first port label.
        :param str node2_id: The second node identifier.
        :param str port2_label: The second port label.
        """
        link_id = Link.calc_id(node1_id, port1_label, node2_id, port2_label)
        self.unlink(link_id)

    def set_link(self, node1_id, port1_label, node2_id, port2_label):
        """
        Set a link between two nodes.

        :param str node1_id: The first node identifier.
        :param str port1_label: The first port label.
        :param str node2_id: The second node identifier.
        :param str port2_label: The second port label.
        """
        link_id = Link.calc_id(node1_id, port1_label, node2_id, port2_label)
        self.relink(link_id)

    def _set_test_log(self, log):
        """
        Set the current test execution log. This log is set by testlog library

        :param log: the logging object requested
        """
        for enode in self.nodes.values():
            enode._set_test_log(log)


__all__ = ['TopologyManager']



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Topology Graph Link module.
"""

from copy import deepcopy

from pprintpp import pformat

from .node import Node
from .port import Port


class Link:
    """
    Topology Graph Link.

    :param str node1: The first node of the link.
    :param str node2: The second node of the link.
    :param str port1: The first port of the link.
    :param str port2: The second port of the link.
    :param dict metadata: The metadata of the link.
    """

    def __init__(
        self,
        node1: Node,
        port1: Port,
        node2: Node,
        port2: Port,
        metadata: dict = None
    ):
        self._node1 = node1
        self._node2 = node2
        self._port1 = port1
        self._port2 = port2
        self._metadata = {} if metadata is None else metadata

        # Allow to override the identifier of the link.
        self.identifier = self.metadata.pop(
            'identifier',
            Link.calc_id(
                node1.identifier,
                port1.label,
                node2.identifier,
                port2.label
            )
        )
        self.attrs = {}

    def as_dict(self) -> dict:
        """
        Returns the link as a dictionary.
        """
        return {
            'node1': self._node1.as_dict(),
            'port1': self._port1.as_dict(),
            'node2': self._node2.as_dict(),
            'port2': self._port2.as_dict(),
            'metadata': deepcopy(self._metadata)
        }

    def __str__(self):
        return pformat(self.as_dict())

    @classmethod
    def calc_id(
        cls,
        node1_id: str,
        port1_label: str,
        node2_id: str,
        port2_label: str
    ) -> str:
        """
        Calculates the identifier of the link.
        :param str node1_id: The identifier of the first node.
        :param str port1_label: The label of the first port.
        :param str node2_id: The identifier of the second node.
        :param str port2_label: The label of the second port.
        """
        port1_id = Port.calc_id(node1_id, port1_label)
        port2_id = Port.calc_id(node2_id, port2_label)

        # Any pair of ports produce the same identifier regardless their order.
        return (
            f'{port1_id} -- {port2_id}'
            if port1_id <= port2_id else
            f'{port2_id} -- {port1_id}'
        )

    @property
    def node1(self) -> Node:
        """
        Returns the first node of the link.
        """
        return self._node1

    @property
    def node2(self) -> Node:
        """
        Returns the second node of the link.
        """
        return self._node2

    @property
    def port1(self) -> Port:
        """
        Returns the first port of the link.
        """
        return self._port1

    @property
    def port2(self) -> Port:
        """
        Returns the second port of the link.
        """
        return self._port2

    @property
    def metadata(self) -> dict:
        """
        Returns a copy of the metadata of the link.
        """
        return deepcopy(self._metadata)

    def has_node(self, node_id: str) -> bool:
        """
        Returns True if the link has the node.

        :param str node_id: The identifier of the node.
        """
        return (
            self.node1.identifier == node_id or
            self.node2.identifier == node_id
        )

    def has_port(self, node_id: str, port_label: str) -> bool:
        """
        Returns True if the link has the port.

        :param str node_id: The identifier of the node.
        :param str port_label: The port label.
        """
        return (
            self.port1.identifier == Port.calc_id(node_id, port_label) or
            self.port2.identifier == Port.calc_id(node_id, port_label)
        )


__all__ = ['Link']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2022 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from pathlib import Path

from setuptools import setup, find_packages


def check_directory():
    """
    You must always change directory to the parent of this file before
    executing the setup.py script. setuptools will fail reading files,
    including and excluding files from the MANIFEST.in, defining the library
    path, etc, if not.
    """
    from os import chdir

    here = Path(__file__).parent.resolve()
    if Path.cwd().resolve() != here:
        print('Changing path to {}'.format(here))
        chdir(str(here))


check_directory()


def read(filename):
    """
    Read a file relative to setup.py location.
    """
    import os
    here = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(here, filename)) as fd:
        return fd.read()


def find_version(filename):
    """
    Find package version in file.
    """
    import re
    content = read(filename)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


def find_requirements(filename):
    """
    Find requirements in file.
    """
    import string
    content = read(filename)
    requirements = []
    for line in content.splitlines():
        line = line.strip()
        if line and line[:1] in string.ascii_letters:
            requirements.append(line)
    return requirements


setup(
    name='topology',
    version=find_version('lib/topology/__init__.py'),
    package_dir={'': 'lib'},
    packages=find_packages('lib'),

    # Dependencies
    install_requires=find_requirements('requirements.txt'),

    # Metadata
    author='Hewlett Packard Enterprise Development LP',
    author_email='hpe-networking@lists.hp.com',
    description=(
        'Topology is a framework for building and testing network '
        'topologies, with support for pytest.'
    ),
    long_description=read('README.rst'),
    long_description_content_type='text/x-rst',
    url='http://topology.rtfd.org/',
    keywords='topology',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    # Entry points
    entry_points={
        'console_scripts': [
            'topology=topology.__main__:main',
        ],
        'pytest11': [
            'topology = topology.pytest.plugin',
        ],
        'topology_platform_10': [
            'debug = topology.platforms.debug:DebugPlatform',
        ],
        'topology_library_10': [
            'common = topology.libraries.common',
        ]
    }
)
